/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef pplite_Poly_templ_hh
#define pplite_Poly_templ_hh

#include "Poly.hh"
#include "Poly_min.hh"

#include <algorithm>
#include <functional>

namespace pplite {

namespace detail {

template <typename Dst, typename Src>
void
concat_sys(Dst& dst, Src&& src) {
  // Add sing elements.
  std::move(src.sing_rows.begin(), src.sing_rows.end(),
            std::back_inserter(dst.sing_rows));
  // Add sk elements.
  const dim_type old_sk_size = dst.sk_rows.size();
  std::move(src.sk_rows.begin(), src.sk_rows.end(),
            std::back_inserter(dst.sk_rows));
  // Add ns elements, shifting their indices.
  auto pos = num_rows(dst.ns_rows);
  for (auto i = 0; i < num_rows(src.ns_rows) ; ++i){
    dst.ns_rows.push_back(std::move(src.ns_rows[i]));
    dst.ns_rows[pos++].shift_assign(old_sk_size);
  }
}

template <typename Rows>
inline void
permute_space_dims(Rows& rows, const Dims& cycle, dim_type d) {
  for (auto& r : rows) {
    if (r.space_dim() < d)
      r.set_space_dim(d);
    r.permute_space_dims(cycle);
  }
}

template <typename Rows>
inline void
permute_space_dims(Poly_Impl::Sys<Rows>& sys, const Dims& cycle) {
  if (cycle.empty())
    return;
  dim_type dim = 1 + *(std::max_element(cycle.begin(), cycle.end()));
  permute_space_dims(sys.sing_rows, cycle, dim);
  permute_space_dims(sys.sk_rows, cycle, dim);
  // ns_rows are unaffected.
}

inline void
remove_invalid_lines(Gens& sing_rows) {
  Index_Set tbr = invalid_lines(sing_rows);
  erase_using_sorted_indices(sing_rows, tbr);
}

inline void
remove_invalid_rays(Gens& sk_rows, NS_Rows& ns_rows) {
  Index_Set tbr = invalid_rays(sk_rows);
  if (tbr.empty())
    return;
  erase_using_sorted_indices(sk_rows, tbr);
  for (auto& ns : ns_rows)
    ns.remove_all(tbr.begin(), tbr.end());
  promote_singletons(ns_rows, sk_rows);
}

inline void
remove_invalid_lines_and_rays(Poly_Impl::Sys<Gens>& gs) {
  remove_invalid_lines(gs.sing_rows);
  remove_invalid_rays(gs.sk_rows, gs.ns_rows);
}

} // namespace detail

template <typename Iter>
void
Poly_Impl::remove_space_dims(Iter first, Iter last) {
  using value_type = typename std::iterator_traits<Iter>::value_type;
  static_assert(std::is_same<value_type, dim_type>::value,
                "Iter must have dim_type as its value type");
  assert(std::all_of(first, last,
                     [this](value_type i) { return 0 <= i && i < dim; }));
  assert(std::adjacent_find(first, last,
                            std::greater_equal<value_type>()) == last);
  const dim_type num_rem = std::distance(first, last);
  if (num_rem == 0)
    return;
  const dim_type new_dim = dim - num_rem;
  ensure_valid_gens();
  if (marked_empty()) {
    dim = new_dim;
    return;
  }
  if (new_dim == 0) {
    dim = new_dim;
    set_universe();
    return;
  }

  Sys<Gens> tmp = std::move(gs);
  detail::concat_sys(tmp, std::move(gs_pending));
  // Process (geometric) generators.
  erase_space_dims(tmp.sing_rows, first, last);
  erase_space_dims(tmp.sk_rows, first, last);
  // Remove invalid generators.
  detail::remove_invalid_lines_and_rays(tmp);
  // Use tmp to reinit the polyhedron.
  reinit_with_gens(new_dim, topol, std::move(tmp));
  assert(check_inv());
}

template <typename Iter>
void
Poly_Impl::add_cons(Iter first, Iter last) {
  assert(std::all_of(first, last,
                     [this](const Con& c) {
                       return c.space_dim() <= this->space_dim();
                     }));
  if (marked_empty())
    return;
  if (space_dim() == 0) {
    if (std::any_of(first, last, std::mem_fn(&Con::is_inconsistent)))
      set_empty();
    return;
  }
  ensure_valid_cons();
  auto& eq = cs_pending.sing_rows;
  auto& sk = cs_pending.sk_rows;
  std::copy_if(first, last, std::back_inserter(eq),
               std::mem_fn(&Con::is_equality));
  std::copy_if(first, last, std::back_inserter(sk),
               /* FIXME: not_fn only works using -std=c++1z */
               // std::not_fn(std::mem_fn(&Con::is_equality)));
               [](const Con& c) { return !c.is_equality(); });
  if (marked_min() && !cs_pending.empty())
    set_status(Status::PENDING);
  assert(check_inv());
}

template <typename Iter>
void
Poly_Impl::add_gens(Iter first, Iter last) {
  assert(std::all_of(first, last,
                     [this](const Gen& g) {
                       return g.space_dim() <= this->space_dim();
                     }));
  ensure_valid_gens();
  if (marked_empty()) {
    auto iter = std::find_if(first, last, std::mem_fn(&Gen::is_point));
    assert(iter != last);
    add_gen(*iter);
    add_gens(first, iter);
    ++iter;
    add_gens(iter, last);
    return;
  }
  if (dim == 0)
    return;
  auto& lines = gs_pending.sing_rows;
  auto& sk = gs_pending.sk_rows;
  std::copy_if(first, last, std::back_inserter(lines),
               std::mem_fn(&Gen::is_line));
  std::copy_if(first, last, std::back_inserter(sk),
               /* FIXME: not_fn only works using -std=c++1z */
               // std::not_fn(std::mem_fn(&Gen::is_line)));
               [](const Gen& g) { return !g.is_line(); });
  if (marked_min() && !gs_pending.empty())
    set_status(Status::PENDING);
  assert(check_inv());
}

template <typename System, typename Out, typename Pred>
Out
Poly_Impl::materialize(const System& sys, const System& pending, Out out,
                       bool cons_and_closed, Pred pred) {
  // Copy singular.
  out = std::copy(sys.sing_rows.begin(), sys.sing_rows.end(), out);
  out = std::copy(pending.sing_rows.begin(), pending.sing_rows.end(), out);
  if (cons_and_closed) {
    // Materializing constraints for a CLOSED poly.
    // Copy skel, but not the (strict) positivity.
    out = std::copy_if(sys.sk_rows.begin(), sys.sk_rows.end(),
                       out, pred);
    out = std::copy_if(pending.sk_rows.begin(), pending.sk_rows.end(),
                       out, pred);
    // No need to materialize non-skel.
    return out;
  }
  // Materializing generators of constraints for an NNC poly.
  // Copy skel.
  out = std::copy(sys.sk_rows.begin(), sys.sk_rows.end(), out);
  out = std::copy(pending.sk_rows.begin(), pending.sk_rows.end(), out);
  // Materialize non-skel.
  for (const auto& ns : sys.ns_rows) {
    *out = detail::materialize(ns, sys.sk_rows);
    ++out;
  }
  for (const auto& ns : pending.ns_rows) {
    *out = detail::materialize(ns, pending.sk_rows);
    ++out;
  }
  return out;
}

template <typename Iter>
void
Poly_Impl::unconstrain(Iter first, Iter last) {
  assert(std::all_of(first, last,
                     [this](dim_type d) { return d < space_dim(); }));
  if (first == last)
    return;
  ensure_valid_gens();
  if (marked_empty())
    return;
  auto& lines = gs_pending.sing_rows;
  for (auto i = first; i != last; ++i)
    lines.push_back(line(Var(*i)));
  if (marked_min())
    set_status(Status::PENDING);
  assert(check_inv());
}

} // namespace pplite

#endif // !defined(pplite_Poly_templ_hh)

