/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "U_Poly.hh"

#include <set>
#include <iostream>

namespace pplite {

template <typename PH>
bool
U_Wrap<PH>::check_inv() const {
  std::string reason;
#ifdef NDEBUG
  auto maybe_dump = []() {};
#else // In debugging mode, be noisy.
  auto maybe_dump = [this, &reason]() {
    std::cerr << reason << std::endl;
    this->ascii_dump(std::cerr);
    std::cerr << reason << std::endl;
  };
#endif
  // Checks on info.
  dim_type num_kdims = 0;
  std::set<dim_type> found_kernel_dim;
  for (dim_type sd = space_dim(); sd-- > 0; ) {
    if (is_kernel_dim(sd)) {
      ++num_kdims;
      const auto k_dim = kernel_dim(sd);
      if (k_dim >= kernel.space_dim()) {
        reason = "U_Wrap broken: KERNEL dimension not in kernel";
        maybe_dump();
        return false;
      }
      if (found_kernel_dim.find(k_dim) != found_kernel_dim.end()) {
        reason = "U_Wrap broken: info is not injective";
        maybe_dump();
        return false;
      }
      found_kernel_dim.insert(k_dim);
    }
  }
  if (num_kdims != kernel.space_dim()) {
    reason = "U_Wrap broken: kernel space dimension does not match "
      "the number of KERNEL dimensions";
    maybe_dump();
    return false;
  }
  // Checks on kernel.
  if (is_empty() && kernel.space_dim() > 0) {
    reason = "U_Wrap kernel broken: empty polyhedron has KERNEL dimensions";
    maybe_dump();
    return false;
  }
  for (dim_type i = kernel.space_dim(); i-- > 0; ) {
    if (!kernel.constrains(Var(i))) {
      reason = "U_Wrap kernel broken: kernel has unconstrained dimension";
      maybe_dump();
      return false;
    }
  }
  // All checks passed.
  return true;
}

template <typename PH>
Dims
U_Wrap<PH>::get_inverse_info() const {
  // Build inverse map (kdim -> sdim).
  const auto k_dim = kernel.space_dim();
  Dims inverse(k_dim);
  for (auto i = space_dim(); i-- > 0; ) {
    if (is_kernel_dim(i))
      inverse[kernel_dim(i)] = i;
  }
  return inverse;
}

template <typename PH>
Linear_Expr
U_Wrap<PH>::shell_2_kernel_expr(const Linear_Expr& s_expr) const {
  Linear_Expr k_expr;
  for (dim_type i = 0, i_end = s_expr.space_dim(); i != i_end; ++i) {
    if (is_kernel_dim(i))
      k_expr.set(kernel_dim(i), s_expr.get(i));
  }
  return k_expr;
}

template <typename PH>
inline Con
U_Wrap<PH>::shell_2_kernel_con(const Con& s_c) const {
  Linear_Expr k_expr = shell_2_kernel_expr(s_c.linear_expr());
  return Con(std::move(k_expr), s_c.inhomo_term(), s_c.type());
}

template <typename PH>
inline Gen
U_Wrap<PH>::shell_2_kernel_gen(const Gen& s_g) const {
  // FIXME: what if invalid line/ray?
  Linear_Expr k_expr = shell_2_kernel_expr(s_g.linear_expr());
  return Gen(s_g.type(), std::move(k_expr), s_g.impl().inhomo);
}

template <typename PH>
void
U_Wrap<PH>::check_kernel_for_unconstrained() const {
  // Note: all changes preserve semantics.
  auto& x = const_cast<U_Wrap&>(*this);
  Vars_Set k_remove;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd)) {
      const Var k_var(x.kernel_dim(sd));
      if (!x.kernel.constrains(k_var)) {
        k_remove.insert(k_var);
        x.info[sd] = not_a_dim();
      }
    }
  }
  x.kernel.remove_space_dims(k_remove);
  x.kernel.minimize();
  // Fix info to match the removed kernel dims (one at a time).
  dim_type found = 0;
  for (auto kd : k_remove) {
    x.decr_info_helper(kd - found);
    ++found;
  }
}

template <typename PH>
Con
U_Wrap<PH>::kernel_2_shell_con(const Con& k_c,
                               const Dims& inverse) const {
  const auto& k_expr = k_c.linear_expr();
  Linear_Expr s_expr;
  for (auto j = k_expr.space_dim(); j-- > 0; ) {
    if (!k_expr[j].is_zero())
      s_expr.set(Var(inverse[j]), k_expr[j]);
  }
  return Con(std::move(s_expr), k_c.inhomo_term(), k_c.type());
}

template <typename PH>
Cons
U_Wrap<PH>::copy_cons() const {
  Cons result;
  Dims inverse = get_inverse_info();
  for (const auto& k_c : kernel.cons())
    result.push_back(kernel_2_shell_con(k_c, inverse));
  return result;
}

template <typename PH>
Gen
U_Wrap<PH>::kernel_2_shell_gen(const Gen& k_g,
                               const Dims& inverse) const {
  Linear_Expr s_expr;
  const auto& k_expr = k_g.linear_expr();
  for (auto i = 0, i_end = k_expr.space_dim(); i != i_end; ++i) {
    if (!k_expr[i].is_zero())
      s_expr.set(Var(inverse[i]), k_expr[i]);
  }
  return Gen(k_g.type(), std::move(s_expr), k_g.impl().inhomo);
}

template <typename PH>
Gens
U_Wrap<PH>::copy_gens() const {
  Gens result;
  if (is_empty())
    return result;
  for (dim_type i = 0, i_end = space_dim(); i != i_end; ++i) {
    if (!is_kernel_dim(i))
      result.push_back(line(Var(i)));
  }
  Dims inverse = get_inverse_info();
  for (const auto& k_g : kernel.gens())
    result.push_back(kernel_2_shell_gen(k_g, inverse));
  return result;
}

template <typename PH>
void
U_Wrap<PH>::sync_kernel_dims(const U_Wrap& y) const {
  // Note: all changes preserve semantics.
  U_Wrap& x = const_cast<U_Wrap&>(*this);
  assert(x.space_dim() == y.space_dim());
  assert(x.kernel.space_dim() == y.kernel.space_dim());
  // Permute the kernel dims of x to match those of y.
  Dims mapper(x.kernel.space_dim(), not_a_dim());
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd)) {
      assert(y.is_kernel_dim(sd));
      mapper[x.kernel_dim(sd)] = y.kernel_dim(sd);
      x.kernel_dim(sd) = y.kernel_dim(sd);
    }
  }
  assert(x.info == y.info);
  x.kernel.map_space_dims(mapper);
}

template <typename PH>
Poly_Con_Rel
U_Wrap<PH>::relation_with(const Con& c) const {
  assert(space_dim() >= c.space_dim());

  if (is_empty())
    return Poly_Con_Rel::saturates()
      && Poly_Con_Rel::is_included()
      && Poly_Con_Rel::is_disjoint();
  if (space_dim() == 0)
    return kernel.relation_with(c);
  check_kernel_for_unconstrained();
  if (!all_kernel_dims(c.linear_expr()))
    return Poly_Con_Rel::strictly_intersects();
  return kernel.relation_with(shell_2_kernel_con(c));
}

template <typename PH>
Poly_Gen_Rel
U_Wrap<PH>::relation_with(const Gen& g) const {
  assert(space_dim() >= g.space_dim());
  if (is_empty())
    return Poly_Gen_Rel::nothing();
  if (space_dim() == 0)
    return Poly_Gen_Rel::subsumes();
  if (g.is_line() || g.is_ray()) {
    // Quick check for subsumption.
    bool quick_subsumes = true;
    for (auto i = g.space_dim(); i-- > 0; ) {
      if (is_kernel_dim(i) && g.coeff(Var(i)) != 0) {
        quick_subsumes = false;
        break;
      }
    }
    if (quick_subsumes)
      return Poly_Gen_Rel::subsumes();
  }
  return kernel.relation_with(shell_2_kernel_gen(g));
}

template <typename PH>
bool
U_Wrap<PH>::equals(const U_Wrap& y) const {
  const auto& x = *this;
  if (x.kernel.space_dim() != y.kernel.space_dim())
    return false;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd) ^ y.is_kernel_dim(sd))
      return false;
  }
  x.sync_kernel_dims(y);
  return x.kernel.equals(y.kernel);
}

template <typename PH>
bool
U_Wrap<PH>::contains(const U_Wrap& y) const {
  const auto& x = *this;
  if (x.is_empty())
    return y.is_empty();
  if (y.is_empty() || x.is_universe())
    return true;

  Vars_Set y_uncon;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd) && y.is_unconstrained_dim(sd))
      return false;
    if (x.is_unconstrained_dim(sd) && y.is_kernel_dim(sd))
      y_uncon.set(sd);
  }

  if (y_uncon.size() == y.kernel.space_dim())
    return false;

  if (y_uncon.empty()) {
    x.sync_kernel_dims(y);
    return x.kernel.contains(y.kernel);
  }

  // Work on a copy of y.
  auto y_copy(y);
  y_copy.unconstrain_kernel(y_uncon);
  assert(x.kernel.space_dim() == y_copy.kernel.space_dim());
  x.sync_kernel_dims(y_copy);
  return x.kernel.contains(y_copy.kernel);
}

template <typename PH>
bool
U_Wrap<PH>::strictly_contains(const U_Wrap& y) const {
  const auto& x = *this;
  if (x.is_empty())
    return false;
  if (y.is_empty())
    return true;
  if (x.is_universe() && !y.is_universe())
    return true;

  Vars_Set y_uncon;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd) && y.is_unconstrained_dim(sd))
      return false;
    if (x.is_unconstrained_dim(sd) && y.is_kernel_dim(sd))
      y_uncon.set(sd);
  }

  if (y_uncon.size() == y.kernel.space_dim())
    return false;

  if (y_uncon.empty()) {
    x.sync_kernel_dims(y);
    return x.kernel.strictly_contains(y.kernel);
  }

  // Work on a copy of y.
  auto y_copy(y);
  y_copy.unconstrain_kernel(y_uncon);
  assert(x.kernel.space_dim() == y_copy.kernel.space_dim());
  x.sync_kernel_dims(y_copy);
  // Note: since we unconstrained some dims of y, we answer positively
  // even if kernel containment is non-strict.
  return x.kernel.contains(y_copy.kernel);
}

template <typename PH>
bool
U_Wrap<PH>::is_disjoint_from(const U_Wrap& y) const {
  const auto& x = *this;
  // No need to consider dims that are unconstrained in both args.
  Vars_Set x_uncon;
  Vars_Set y_uncon;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (x.is_kernel_dim(sd)) {
      if (!y.is_kernel_dim(sd))
        x_uncon.set(sd);
    }
    else if (y.is_kernel_dim(sd)) {
      y_uncon.set(sd);
    }
  }

  if (x_uncon.empty()) {
    if (y_uncon.empty()) {
      x.sync_kernel_dims(y);
      return x.kernel.is_disjoint_from(y.kernel);
    }
    auto y_copy(y);
    y_copy.unconstrain(y_uncon);
    x.sync_kernel_dims(y_copy);
    return x.kernel.is_disjoint_from(y_copy.kernel);
  }

  auto x_copy(x);
  x_copy.unconstrain(x_uncon);
  if (y_uncon.empty()) {
    x_copy.sync_kernel_dims(y);
    return x_copy.kernel.is_disjoint_from(y.kernel);
  }
  auto y_copy(y);
  y_copy.unconstrain(y_uncon);
  x_copy.sync_kernel_dims(y_copy);
  return x_copy.kernel.is_disjoint_from(y_copy.kernel);
}

template <typename PH>
void
U_Wrap<PH>::intersection_assign(const U_Wrap& y) {
  auto& x = *this;
  if (x.is_empty() || y.is_universe())
    return;
  if (y.is_empty() || x.is_universe()) {
    x = y;
    return;
  }

  // Both non-empty and non-universe (hence, non zero-dim).
  assert(x.kernel.space_dim() > 0 && y.kernel.space_dim() > 0);

  // If a dimension is constrained in any of the two arguments,
  // it will be so in the result too.
  const dim_type x_old_k_dim = x.kernel.space_dim();
  dim_type x_new_k_dims = 0;
  Vars_Set y_new_kvars;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (!x.is_kernel_dim(sd)) {
      if (y.is_kernel_dim(sd)) {
        x.info[sd] = x_old_k_dim + x_new_k_dims;
        ++x_new_k_dims;
      }
      continue;
    }
    if (!y.is_kernel_dim(sd))
      y_new_kvars.set(sd);
  }

  x.kernel.add_space_dims(x_new_k_dims);
  if (y_new_kvars.empty()) {
    // No need to copy&modify y.kernel.
    x.sync_kernel_dims(y);
    x.kernel.intersection_assign(y.kernel);
  } else {
    // Work on a copy of y.
    auto y_copy(y);
    const dim_type y_old_k_dim = y_copy.kernel.space_dim();
    dim_type y_new_k_dims = 0;
    for (auto j : y_new_kvars) {
      y_copy.info[j] = y_old_k_dim + y_new_k_dims;
      ++y_new_k_dims;
    }
    y_copy.kernel.add_space_dims(y_new_k_dims);
    x.sync_kernel_dims(y_copy);
    x.kernel.intersection_assign(y_copy.kernel);
  }
  if (x.kernel.is_empty())
    x.set_empty();

  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::poly_hull_assign(const U_Wrap& y) {
  auto& x = *this;
  if (y.is_empty() || x.is_universe())
    return;
  if (x.is_empty() || y.is_universe()) {
    x = y;
    return;
  }
  assert(x.kernel.space_dim() > 0 && y.kernel.space_dim() > 0);
  Vars_Set x_uncon;
  Vars_Set y_uncon;
  // If a dim is unconstrained in any of the two arguments,
  // it will be so in the result too.
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (!x.is_kernel_dim(sd)) {
      if (y.is_kernel_dim(sd))
        y_uncon.set(sd);
    }
    else if (!y.is_kernel_dim(sd)) {
      x_uncon.set(sd);
    }
  }

  if (y_uncon.size() == y.kernel.space_dim()) {
    // y would become universe, hence also x.
    x.set_universe();
    return;
  }

  if (!x_uncon.empty()) {
    // Refrain from directly calling `unconstrain'
    // (since it may lead to kernels having different dimensions).
    x.unconstrain_kernel(x_uncon);
    // Check if we unconstrained all kernel dims.
    if (x.kernel.is_universe()) {
      x.set_universe();
      return;
    }
  }

  if (y_uncon.empty()) {
    x.sync_kernel_dims(y);
    x.kernel.poly_hull_assign(y.kernel);
    x.check_kernel_for_unconstrained();
    return;
  }

  // Work on a copy of y.
  auto y_copy(y);
  // Refrain from directly calling `unconstrain' (ditto).
  y_copy.unconstrain_kernel(y_uncon);
  if (y_copy.kernel.is_universe()) {
    x.set_universe();
    return;
  }

  x.sync_kernel_dims(y_copy);
  x.kernel.poly_hull_assign(y_copy.kernel);
  x.check_kernel_for_unconstrained();
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::time_elapse_assign(const U_Wrap& y) {
  auto& x = *this;
  if (x.is_empty() || y.is_empty()) {
    x.set_empty();
    return;
  }
  if (x.is_universe() || y.is_universe()) {
    x.set_universe();
    return;
  }

  Index_Set uncon;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (y.is_unconstrained_dim(sd) && x.is_kernel_dim(sd))
      uncon.set(sd);
  }
  x.unconstrain(uncon);
  if (x.is_universe())
    return;

  Gens y_s_gens;
  Dims y_inv = y.get_inverse_info();
  for (const auto& k_g : y.kernel.gens())
    y_s_gens.push_back(y.kernel_2_shell_gen(k_g, y_inv));

  Gens rays;
  detail::add_as_rays(std::move(y_s_gens), rays);
  x.add_gens(std::move(rays));
  assert(x.check_inv());
}

template <typename PH>
void
U_Wrap<PH>::widening_assign(const U_Wrap& y, widen_type widen,
                            const Cons* upto_ptr) {
  auto& x = *this;
  // Filter out empty and zero dim cases.
  if (x.space_dim() == 0 || x.is_empty() || y.is_empty())
    return;
  // Filter out x universe case.
  if (x.is_universe())
    return;

  Index_Set upto;
  if (upto_ptr) {
    const auto& upto_cs = *upto_ptr;
    for (dim_type i = upto_cs.size(); i-- > 0; ) {
      if (x.relation_with(upto_cs[i]).implies(Poly_Con_Rel::is_included()))
        upto.set(i);
    }
  }

  const dim_type x_old_k_dim = x.kernel.space_dim();
  dim_type x_new_k_dims = 0;
  for (dim_type sd = x.space_dim(); sd-- > 0; ) {
    if (y.is_unconstrained_dim(sd))
      assert(x.is_unconstrained_dim(sd));
    else if (x.is_unconstrained_dim(sd)) {
      x.info[sd] = x_old_k_dim + x_new_k_dims;
      ++x_new_k_dims;
    }
  }
  x.kernel.add_space_dims(x_new_k_dims);
  x.sync_kernel_dims(y);
  x.kernel.widening_assign(y.kernel, widen);
  x.check_kernel_for_unconstrained();

  if (upto.size() > 0) {
    for (auto i : upto)
      x.add_con((*upto_ptr)[i]);
  }
  assert(x.check_inv());
}

template <typename PH>
void
U_Wrap<PH>::concatenate_assign(const U_Wrap& y) {
  auto& x = *this;
  const dim_type old_sdim = x.space_dim();
  x.info.insert(x.info.end(), y.info.begin(), y.info.end());
  if (x.is_empty() || y.is_empty()) {
    x.set_empty();
    return;
  }
  // Adjust concatenated info (if needed).
  const dim_type old_kdim = x.kernel.space_dim();
  if (old_kdim > 0 && y.kernel.space_dim() > 0) {
    for (dim_type i = old_sdim, i_end = x.space_dim(); i != i_end; ++i) {
      if (x.is_kernel_dim(i))
        x.kernel_dim(i) += old_kdim;
    }
  }
  // Concatenate (non-empty) kernels.
  x.kernel.concatenate_assign(y.kernel);
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::add_space_dims(dim_type m, bool project) {
  if (m == 0)
    return;
  const dim_type new_dim = space_dim() + m;
  if (!project || is_empty()) {
    info.resize(new_dim, not_a_dim());
    assert(check_inv());
    return;
  }
  const dim_type kd = kernel.space_dim();
  info.reserve(new_dim);
  for (dim_type i = 0; i < m; ++i)
    info.push_back(kd + i);
  kernel.add_space_dims(m, project);
  // FIXME: needed? kernel.minimize();
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::map_space_dims(const Dims& pfunc) {
  if (space_dim() == 0)
    return;
  assert(space_dim() == num_rows(pfunc));
  const auto max_dim = *std::max_element(pfunc.begin(), pfunc.end());
  if (max_dim == not_a_dim()) {
    // Empty codomain.
    info.clear();
    if (kernel.is_empty())
      set_empty();
    else
      set_universe();
    return;
  }

  const auto old_s_dim = space_dim();
  const auto new_s_dim = max_dim + 1;
  // First, unconstrain those kernel variables that will be removed.
  Vars_Set uncon;
  for (dim_type i = old_s_dim; i-- > 0; ) {
    if (is_kernel_dim(i) && pfunc[i] == not_a_dim())
      uncon.set(i);
  }
  unconstrain(uncon);
  // Then, compute new shell-kernel mapping.
  Dims new_info(new_s_dim, not_a_dim());
  for (dim_type i = old_s_dim; i-- > 0; ) {
    const auto j = pfunc[i];
    if (j != not_a_dim())
      new_info[j] = info[i];
  }
  using std::swap;
  swap(info, new_info);
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::expand_space_dim(Var v, dim_type m) {
  if (m == 0)
    return;
  assert(m > 0);
  const dim_type sd = v.id();
  const dim_type new_sdim = space_dim() + m;
  if (is_unconstrained_dim(sd) || is_empty()) {
    info.resize(new_sdim, not_a_dim());
    assert(check_inv());
    return;
  }
  const dim_type old_kdim = kernel.space_dim();
  info.reserve(new_sdim);
  for (dim_type i = 0; i < m; ++i)
    info.push_back(old_kdim + i);
  Var kv(kernel_dim(sd));
  kernel.expand_space_dim(kv, m);
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::fold_space_dims(const Index_Set& vars, Var dest) {
  if (is_unconstrained_dim(dest.id())) {
    remove_space_dims(vars);
    return;
  }
  if (std::any_of(vars.begin(), vars.end(),
                  [this](dim_type i) { return is_unconstrained_dim(i); })) {
    // vars contains an unconstrained dimension.
    remove_space_dims(vars);
    unconstrain(dest);
    assert(check_inv());
  }
  // All dims (in vars and dest) are constrained.
  Var kdest(kernel_dim(dest.id()));
  Index_Set kvars;
  for (auto i : vars) {
    kvars.set(kernel_dim(i));
    info[i] = not_a_dim();
  }
  erase_using_sorted_indices(info, vars);
  kernel.fold_space_dims(kvars, kdest);
  check_kernel_for_unconstrained();
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::affine_image_aux(bool preimage,
                             Var var, const Linear_Expr& expr,
                             const Integer& inhomo, const Integer& den) {
  if (is_empty())
    return;
  if (is_unconstrained_dim(var.id()) && expr.get(var) != 0)
    // Invertible affine (pre-)image on an unconstrained variable: nop.
    return;

  // Update info with dimensions that will enter the kernel.
  const dim_type old_k_dim = kernel.space_dim();
  dim_type new_k_dims = 0;
  if (is_unconstrained_dim(var.id())) {
    info[var.id()] = old_k_dim + new_k_dims;
    ++new_k_dims;
  }
  for (dim_type j = 0, j_end = expr.space_dim(); j != j_end; ++j) {
    if (is_unconstrained_dim(j) && expr.get(j) != 0) {
      assert(j != var.id());
      info[j] = old_k_dim + new_k_dims;
      ++new_k_dims;
    }
  }
  kernel.add_space_dims(new_k_dims);
  // Rewrite var and expr for kernel.
  Var k_var(kernel_dim(var.id()));
  Linear_Expr k_expr = shell_2_kernel_expr(expr);
  // Apply affine (pre-)image on kernel.
  if (preimage)
    kernel.affine_preimage(k_var, k_expr, inhomo, den);
  else
    kernel.affine_image(k_var, k_expr, inhomo, den);
  check_kernel_for_unconstrained();
  assert(check_inv());
}

template <typename PH>
void
U_Wrap<PH>::parallel_affine_image(Vars& vars,
                                  Linear_Exprs& exprs,
                                  Integers& inhomos, Integers& dens) {
  const auto nvars = num_rows(vars);
  if (nvars == 1) {
    affine_image(vars[0], exprs[0], inhomos[0], dens[0]);
    return;
  }
  if (is_empty())
    return;

  // Update info with dimensions that will enter the kernel.
  const dim_type old_k_dim = kernel.space_dim();
  dim_type new_k_dims = 0;
  for (dim_type i = 0; i != nvars; ++i) {
    const auto sd = vars[i].id();
    if (is_unconstrained_dim(sd)) {
      info[sd] = old_k_dim + new_k_dims;
      ++new_k_dims;
    }
    const auto& e_i = exprs[i];
    for (dim_type j = 0, j_end = e_i.space_dim(); j != j_end; ++j) {
      if (is_kernel_dim(j) || e_i.get(j).is_zero())
        continue;
      info[j] = old_k_dim + new_k_dims;
      ++new_k_dims;
    }
  }
  kernel.add_space_dims(new_k_dims);
  // Rewrite vars and exprs for kernel.
  for (dim_type i = 0; i != nvars; ++i) {
    const auto sd = vars[i].id();
    vars[i] = Var(kernel_dim(sd));
    exprs[i] = shell_2_kernel_expr(exprs[i]);
  }
  // Apply parallel affine image on kernel.
  kernel.parallel_affine_image(vars, exprs, inhomos, dens);
  check_kernel_for_unconstrained();
  assert(check_inv());
}

template <typename PH>
bool
U_Wrap<PH>::min_max(bool minimizing,
                    const Linear_Expr& expr, const Integer& inhomo,
                    Integer& num, Integer& den, bool& included,
                    Gen& g) const {
  if (is_empty())
    return false;
  Linear_Expr k_expr;
  for (dim_type j = 0, j_end = expr.space_dim(); j != j_end; ++j) {
    if (expr.get(j) == 0)
      continue;
    if (is_unconstrained_dim(j))
      return false;
    k_expr.set(kernel_dim(j), expr.get(j));
  }
  Gen k_g;
  if (kernel.min_max(minimizing, k_expr, inhomo, num, den, included, k_g)) {
    Dims inverse = get_inverse_info();
    g = kernel_2_shell_gen(k_g, inverse);
    return true;
  }
  return false;
}

template <typename PH>
bool
U_Wrap<PH>::ascii_load(std::istream& s) {
  std::string str;
  dim_type s_dim;
  if (not (ascii_load_string(s, "space_dim") && (s >> s_dim) && (s_dim >= 0)))
    return false;
  if (!(s >> str) || (str != "empty" && str != "not_empty"))
    return false;
  bool flag_empty = (str == "empty");
  info.resize(s_dim, not_a_dim());
  for (dim_type i = 0; i < s_dim; ++i) {
    dim_type j;
    if (not (ascii_load_string(s, "dim")))
      return false;
    if (!(s >> j) || j != i)
      return false;
    if (not (ascii_load_string(s, "->")))
      return false;
    if (!(s >> str))
      return false;
    if (str != "not_a_dim")
      info[i] = std::stol(str);
  }
  if (!kernel.ascii_load(s))
    return false;
  if (kernel.is_empty() != flag_empty)
    return false;
  // Check invariants.
  assert(check_inv());
  return true;
}

template <typename PH>
void
U_Wrap<PH>::ascii_dump(std::ostream& s) const {
  s << "space_dim " << space_dim();
  s << " " << (is_empty() ? "empty" : "not_empty") << "\n";
  for (unsigned i = 0, i_end = info.size(); i != i_end; ++i) {
    s << "dim " << i << " -> ";
    if (is_kernel_dim(i))
      s << kernel_dim(i);
    else
      s << "not_a_dim";
    s << "\n";
  }
  s << "\n";
  kernel.ascii_dump(s);
  s << "\n";
}

// Explicit template instantiation (definition).
template class U_Wrap<Poly>;

} // namespace pplite
